package dev.pfaff.altargui.util;

import org.jspecify.annotations.Nullable;

import java.lang.invoke.MethodHandles;
import java.lang.invoke.VarHandle;
import java.util.ArrayList;
import java.util.concurrent.locks.LockSupport;

public final class Trigger {
	private static final VarHandle VH_triggered;
	private static final VarHandle VH_head;

	static {
		try {
			VH_triggered = MethodHandles.lookup()
					.findVarHandle(Trigger.class, "triggered", boolean.class)
					.withInvokeExactBehavior();
			VH_head = MethodHandles.lookup()
					.findVarHandle(Trigger.class, "head", Node.class)
					.withInvokeExactBehavior();
		} catch (NoSuchFieldException | IllegalAccessException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Initially not triggered.
	 */
	public Trigger() {
	}

	public Trigger(boolean initiallyTriggered) {
		this.triggered = initiallyTriggered;
	}

	private boolean triggered;
	private @Nullable Node head;

	public void trigger() {
		VH_triggered.set(this, true);
		var node = (Node) VH_head.getAndSet(this, (Node) null);
		while (node != null) {
			LockSupport.unpark(node.waiter);
			node = node.next;
		}
	}

	public void reset() {
		var node = (Node) VH_head.getAndSet(this, (Node) null);
		VH_triggered.set(this, false);
		if (node != null) {
			var list = new ArrayList<Thread>();
			while (node != null) {
				list.add(node.waiter);
				node = node.next;
			}
			throw new IllegalStateException(STR."There are \{list.size()} threads waiting on the trigger upon reset: \{list}");
		}
	}

	public void await() {
		if (!(boolean) VH_triggered.getAcquire(this)) {
			var thread = Thread.currentThread();
			Node head;
			do {
				head = (Node) VH_head.get(this);
			} while (!VH_head.compareAndSet(this, head, new Node(thread, head)));
			LockSupport.park();
		}
	}

	private record Node(Thread waiter, @Nullable Node next) {}
}
